// This is -*- C++ -*-
// $Id: AsciiImport.cpp,v 1.5 1998/12/15 20:56:28 trow Exp $

/* 
 * AsciiImport.cpp
 *
 * Copyright (C) 1998 Asger Alstrup Nielsen
 *
 * Developed by Asger Alstrup Nielsen <alstrup@diku.dk>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <config.h>
#include "AsciiImport.h"

void AsciiImport::autosetup(int learn_lines) {
  if (heading_lines() == -1) {
    autosetup_heading_lines();
  }

  // Now, skip the headings while keeping track of the maximum width
  ascii_reader().restart(true);
  paperwidth = 0;
  for (int i=0; i < heading_lines(); ++i) {
    size_t l = string_tool.expand_tabs(ascii_reader().readline(), 
				       tab_size()).length();
    if (l == 0) return;
    if (l > paperwidth) paperwidth = l;
  }

  // Then collect the learning material
  learning_lines.clear();
  for (int i=0; i < learn_lines; ++i) {
    string line = ascii_reader().readline();
    if (line.empty()) break;
    size_t width = string_tool.expand_tabs(line, tab_size()).length();
    if (width > paperwidth) paperwidth = width;
    learning_lines.push_back(line);
  }

  // Now, determine and setup the kind of column divider we have
  if (column_divider_ == 0) {
    autosetup_column_divider();
  } else {
    column_divider().autosetup(learning_lines, paperwidth);
  }
  // Set up the data types
  autosetup_datatypes();
  // And finally, set up the data sets
  autosetup_datasets();
}

// This imports the data
void AsciiImport::import(int lines) {
  ascii_reader().restart(lines != -1);

  // Reset performance results
  missing_ = 0;
  errors_ = 0;
  correct_ = 0;

  // First, we create the empty headings
  headings().clear();
  while (headings().size() <= column_divider().columns()) {
    headings().push_back(string());
  }

  // Then we read the headings
  for (int i=0; i < heading_lines(); ++i) {
    string line = ascii_reader().readline();
    if (line.empty())
      break;
    string_tools::strings cols = column_divider().split(line);
    string superflous = checkSize(cols);
    for (size_t j=0; j < cols.size(); ++j) {
      headings()[j] = string_tool.concatenate_words(headings()[j], cols[j]);
      // If there was superflous text, add it into the heading
      if (!superflous.empty() && j == cols.size() - 1) {
	headings()[j] = string_tool.concatenate_words(headings()[j], superflous);
      }
    }
  }

  // And then the data
  int l = 0;
  while (l < lines || lines == -1) {
    // Get a row
    string line = ascii_reader().readline();
    if (line.empty())
      break;
    ++l;

    // Split it up
    string_tools::strings cols = column_divider().split(line);
    string superflous = checkSize(cols);
    if (!superflous.empty()) {
      ++errors_;
      warning_controller().warn(string("Ignoring superflous text '") + superflous + string("'"));
    }

    // And feed it into the container
    for (size_t i=0; i < cols.size(); ++i) {
      Value * v = 0;
      if (cols[i].empty()) {
	++missing_;
      } else {
	string super;
	v = datatypes()[i]->parse(cols[i], super);
	if (super == cols[i]) {
	  ++errors_;
	  warning_controller().warn(string("Error parsing string '") + cols[i] + string("'"));
	} else if (!super.empty()) {
	  warning_controller().warn(string("Superflous chars '") + super + 
				    string("' parsing string '") + cols[i] + string("'"));
	  // We count it as a half correct
	  ++correct_;
	} else {
	  // A correct gets two points
	  ++correct_;
	  ++correct_;
	}
      }
      if (v == 0 && !datasets()[i]->support_missing()) {
	// If we didn't get any value, and do not support missing data,
	// we have to provide a default value
	v = datatypes()[i]->create_default_value();
	warning_controller().warn(string("Using default value in row ") 
				  + string_tool.int2str(datasets()
							[i]->size() + 1) 
				  + string(" column ")
				  + string_tool.int2str(i)
				  + string(" because of missing data."));
      }
      datasets()[i]->push_back(v);
      delete v;
    }
  }
}


string AsciiImport::checkSize(string_tools::strings & result) {
  string msg;
  size_t size = column_divider().columns();
  // Is the size correct?
  if (result.size() != size) {
    string empty;
    while (result.size() < size) {
      result.push_back(empty);
    }
    while (result.size() > size) {
      string last=result.back();
      result.pop_back();
      if (!msg.empty()) {
	msg = string("' '") + msg;
      }
      msg = last + msg;
    }
  }
  return msg;
}


void AsciiImport::autosetup_heading_lines() {
  // Algorithm:  When the number of numbers in a line
  // is higher than the number of letters, we have data.
  // The characters ',', '.', 'e', 'E', tab and space
  // are judged ambigious, and thus not included in either count.
  ascii_reader().restart(true);

  set_heading_lines(0);
  for (int i=0; i < 10; ++i) {
    string line = ascii_reader().readline();
    if (line.empty()) 
      return;
    size_t count_of_number_chars = 0;
    size_t count_of_heading_chars = 0;
    for (size_t j = 0; j < line.length(); ++j) {
      char buf[2]; buf[0] = line[j]; buf[1] = 0;
      if (string_tool.is_number_char(buf))
	++count_of_number_chars;
      else if (!string_tool.ambigious_number_char(buf)) 
	++count_of_heading_chars;
    }
    if (count_of_number_chars > count_of_heading_chars) {
      set_heading_lines(i);
      return;
    }
  }
  warning_controller().warn("Confused about headings. Assuming 1 line heading.");
  set_heading_lines(1);
}

void AsciiImport::autosetup_column_divider()
{
  // If we have one, get rid of it
  if (column_divider_ != 0) {
    delete column_divider_;
    column_divider_ = 0;
  }

  // We have to decide which kind of divider we want
  // Algorithm:  Try all available kinds, and then
  // choose the one with the best performance
  DataTypes original_datatypes = datatypes();
  Datasets original_datasets = datasets();

  FixedColumns * fixed = new FixedColumns(tab_size());
  fixed->autosetup(learning_lines, paperwidth);
  column_divider_ = fixed;
  autosetup_datatypes();
  autosetup_datasets();
  import(learning_lines.size());
  int score = correct() - errors();
  datatypes_ = original_datatypes;
  datasets_ = original_datasets;
    
  DelimitDivider * delimit = new DelimitDivider();
  delimit->autosetup(learning_lines, paperwidth);
  column_divider_ = delimit;
  autosetup_datatypes();
  autosetup_datasets();
  import(learning_lines.size());
  datatypes_ = original_datatypes;
  datasets_ = original_datasets;

  if (score >= int(correct() - errors())) {
    column_divider_ = fixed;
    delete delimit;
  } else {
    column_divider_ = delimit;
    delete fixed;
  }
}

void AsciiImport::autosetup_datatypes() {
  // TODO: Add logic to choose which kind of data we have in each column
  // For now, we create RealTypes for the lot
  for (size_t i=datatypes().size(); i < column_divider().columns(); ++i) {
    RealType * type = new RealType;
    datatypes().push_back(type);
  }
}


void AsciiImport::autosetup_datasets() {
  // TODO: Add logic to autosetup the Dataset in each column
  // For now, we create real Datasets for the lot
  for (size_t i=datasets().size(); i < column_divider().columns(); ++i) {
    RealSet * data = new RealSet;
    datasets().push_back(data);
  }
}
