// This is -*- C++ -*-
// $Id: Category.cpp,v 1.2 1998/12/15 20:56:30 trow Exp $

/* 
 * Category.cpp
 *
 * Copyright (C) 1998 EMC Capital Management, Inc.
 *
 * Developed by Jon Trowbridge <trow@emccta.com>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <config.h>
#include <stdio.h>
#include "Category.h"

Category_Impl::Category_Impl(catkey_t key) :
  key_(key), locked_(false), ordinal_(false)
{
  char buffer[64];
  sprintf(buffer, "Unnameed #%d", key_+1);
  name_ = string(buffer);
}

Category_Impl::~Category_Impl()
{

}

void
Category_Impl::set_name(const string& l)
{
  if (locked()) throw Exception("Can't change a locked Category");
  name_ = l;
}

void
Category_Impl::add(const string& s)
{
  if (locked()) throw Exception("Can't change a locked Category");
  catcode_t code = (catcode_t)code_lookup_.size();
  pair<string, catcode_t> p(s,code);
  pair<map<string, catcode_t>::iterator,bool> p2 = string_lookup_.insert(p);

  /*
    We are tolerant of collisions, in order to make it easier for category-
    definers.  This way you could, say, iterate over a whole list of strings
    and just add() them all, yielding a Category that reflects the composition
    of that list.
  */
  // assert(p2.second);

  if (p2.second)
    code_lookup_.push_back(s);
}

void
Category_Impl::make_ordinal()
{
  if (locked()) throw Exception("Can't change a locked Category");
  ordinal_ = true;
}

void 
Category_Impl::complete_category()
{
  locked_ = true;
}

catcode_t
Category_Impl::code(const string& s) const
{
  map<string, catcode_t>::const_iterator i = string_lookup_.find(s);
  if (i == string_lookup_.end())
    throw Exception("\""+s+"\" is not part of Category \""+name()+"\"");
  return (*i).second;
}

const string&
Category_Impl::decode(catcode_t c) const
{
  return code_lookup_[c];
}

/////////////////////////////////////////////////////////////////////////////

catkey_t Category::next_key_ = 1;
map<catkey_t, Category_Impl*> Category::impl_map_;
map<string, Category_Impl*> Category::impl_stringmap_;

Category::Category() 
{
  impl_ = new Category_Impl(next_key_++);
}

Category::Category(catkey_t key)
{
  map<catkey_t, Category_Impl*>::const_iterator i = impl_map_.find(key);
  impl_ = (i == impl_map_.end()) ? 0 : (*i).second;
}

Category::Category(const string& l)
{
  map<string, Category_Impl*>::const_iterator i = impl_stringmap_.find(l);
  impl_ = (i == impl_stringmap_.end()) ? 0 : (*i).second;
}

Category::Category(const Category& cat) : impl_(0)
{
  if (! cat.locked())
    throw Exception("Can't copy an unlocked Category");
  
  impl_ = cat.impl_;
}

Category::~Category()
{
  if (impl_ && size() && ! impl_->locked()) {
    cerr << "Deleteing unfinished Category ";
    cerr << '"' << impl_->name() << '"' << endl;
    delete impl_;
  }
}

Category&
Category::operator=(const Category& cat)
{
  if (! cat.locked())
    throw Exception("Can't copy an unlocked Category");

  if (impl_ != &cat) {
    if (! impl_->locked() && size()) {
      cerr << "Overwriting unfinished Category ";
      cerr << '"' << impl_->name() << '"' << endl;
      delete impl_;
    }
    impl_ = cat.impl_;
  }

  return *this;
}

catkey_t
Category::category_key() const
{
  if (!impl_) throw Exception("Attempt to access invalid Category");
  return impl_->category_key();
}

const string&
Category::name() const
{
  if (!impl_) throw Exception("Attempt to access invalid Category");;
  return impl_->name();
}

void
Category::set_name(const string& l)
{
  if (!impl_) throw Exception("Attempt to access invalid Category");;
  impl_->set_name(l);
}

void
Category::add(const string& c)
{
  if (!impl_) throw Exception("Attempt to access invalid Category");;
  impl_->add(c);
}

void
Category::make_ordinal()
{
  if (!impl_) throw Exception("Attempt to access invalid Category");;
  impl_->make_ordinal();
}

void
Category::complete_category()
{
  if (!impl_) throw Exception("Attempt to access invalid Category");;

  if (impl_->locked())
    throw Exception("Attempt to re-complete Category");

  if (impl_->size() < 2)
    throw Exception("A Category must contain at least two elements");

  Category name_collision_check(impl_->name());

  if (name_collision_check.valid())
    throw Exception("Attempt to redefine category \"" + name() + "\"");

  impl_->complete_category();

  impl_map_[impl_->category_key()] = impl_;
  impl_stringmap_[impl_->name()] = impl_;
}

bool
Category::locked() const 
{
  if (!impl_) throw Exception("Attempt to access invalid Category");;
  return impl_->locked();
}

size_t
Category::size() const
{
  if (!impl_) throw Exception("Attempt to access invalid Category");;
  return impl_->size();
}

bool
Category::ordinal() const
{
  if (!impl_) throw Exception("Attempt to access invalid Category");;
  return impl_->ordinal();
}

catcode_t
Category::code(const string& x) const
{
  if (!impl_) throw Exception("Attempt to access invalid Category");;
  return impl_->code(x);
}

const string&
Category::decode(catcode_t c) const
{
  if (!impl_) throw Exception("Attempt to access invalid Category");;
  assert_valid(c);
  return impl_->decode(c);
}




// $Id: Category.cpp,v 1.2 1998/12/15 20:56:30 trow Exp $
