// This is -*- C++ -*-
// $Id: DataSet.h,v 1.28 1998/12/30 02:21:42 trow Exp $

/* 
 * DataSet.h
 *
 * Copyright (C) 1998 Asger Alstrup Nielsen
 *
 * Developed by Asger Alstrup Nielsen <alstrup@diku.dk>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#ifndef _INC_GOOSESET_H
#define _INC_GOOSESET_H

#include <string>
#include "DataType.h"
#include "Value.h"
#include "Random.h"

class Permutation;

/*
 * Base class for container sets.
 *
 * This abstract base class has the responsibility for implementing
 * the DataType and the label, and nothing else.
 */
class DataSet {
public:
  DataSet(DataSet const & s) : type_(0) { assign(s); }
  virtual ~DataSet() { if (type_ != 0) delete type_; }

  virtual DataSet& operator=(DataSet const & s) { assign(s); return *this; }

  // Type control
  DataType * type() { return type_; }
  DataType const * type() const { return type_; }
  void set_type(DataType const & t) {
    if (type_ != 0) delete type_;
    type_ = new DataType(t);
  }

  // Does this set support missing elements in the middle?
  virtual bool support_missing() const { return false; }

  // Label control
  string label() { return label_; }
  string const & label() const { return label_; }
  void set_label(string const & s) { label_ = s; }

  // Access control

  virtual bool empty() const { return size() == 0; }

  // How many values do we have?
  virtual size_t size() const = 0;

  // Notice: no bounds checking
  virtual Value const * operator[](size_t n) const = 0;

  // Make sure we have allocated room for at least N entries
  virtual void reserve(size_t N) = 0;

  // Inserts a new value at the end
  virtual void push_back(Value const * v) = 0;

  // Insert value at position
  virtual void insert(size_t position, Value const * v) = 0;

  // Deletes last element
  virtual void pop_back() = 0;

  // Delete given element
  virtual void erase(size_t position) = 0;

  // Delete elements in range
  virtual void erase(size_t first, size_t last) = 0;

  // Erases all elements
  virtual void clear() = 0;

  // Return a permutation that will, when applied, bring the DataSet into
  // sorted order.  (Does not sort the DataSet, just produces instructions
  // for how to do so!)
  virtual Permutation sorting_permutation(int dir = 1) const = 0;

  // Permute the goose set given a permutation
  virtual void permute(const Permutation&) = 0;

  // Useful for debugging
  virtual void spew(ostream& out, char sep = ' ') const = 0;

  static void random_seed(unsigned s) { rnd_.seed(s); }

protected:
  DataSet() : type_(0) { }
  DataSet(DataType * t) : type_(t) { }

  void assign(DataSet const & s) {
    label_ = s.label_;
    set_type(*s.type_);
  }

  // We cast out const
  Random& random() const { return (Random&)rnd_; }

private:
  string label_;
  DataType * type_;
  static Random rnd_;
};

#endif
