// This is -*- C++ -*-

/* 
 * DataType.h
 *
 * Copyright (C) 1998 EMC Capital Management, Inc.
 *
 * Developed by Havoc Pennington <hp@pobox.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#ifndef _INC_DATATYPE_H
#define _INC_DATATYPE_H

#include <string>
#include "stringtools.h"

class Value;

using namespace std;

// Base class for a type

class DataType {
public:
  enum Kind { Unknown, Real, String };
  DataType() { } 
  DataType(DataType const & dt) { }
  virtual ~DataType() { }
  virtual DataType & operator=(DataType const & dt) { return *this; }
  virtual Kind kind() const { return Unknown; }

  // The GUI name of the type
  virtual string name() const;

  // Format a value according to setup
  virtual string format(Value const & v) const;

  // Parse a string and return the result.
  // We return the amount of the string that was not parsed in
  // 'remains'.
  // The method should try hard to get some useful information
  // out of the string.
  virtual Value * parse(string const & v, string & remains) const { return 0; }

  // Create a dummy value of this type
  virtual Value * create_default_value() const { return 0; }

  // How likely is it that this string is of this type?
  // Returns a "probability" in the range [0,1], where 1.0 signifies
  // that the parse() method will be able to parse it without
  // any error.  0.0 signifies that parse() will not be able to do
  // much with the string.
  virtual double identify(string const & v) const { return 0.0; }

  // Set up parameters for string parsing from a bunch of examples
  virtual void autosetup(string_tools::strings const & learning) { }

  // Virtual copy constructor
  virtual DataType* clone() const { return new DataType(*this); }
};

// A real number
class RealType : public DataType {
public:
  RealType() { }
  RealType(RealType const & t) { }
  virtual ~RealType() { }
  Kind kind() const { return Real; }
  string name () const;
  string format(Value const & v) const;
  Value * parse(string const & v, string & remains) const;
  Value * create_default_value() const;
  double identify(string const & v) const;
  void autosetup(string_tools::strings const & learning);

  virtual DataType* clone() const { return new RealType(*this); }
private:
  // Here, we'll have information about decimal point
  // symbol, thousand separators, and other real number
  // formatting issues.
  // string decimal_point;
  // string thousands_seperator;
  // bool scientific_notation;

  // This should copy any setup information
  void assign(DataType const & d);
};


// A string
class StringType : public DataType {
public:
  StringType() { }
  StringType(StringType const & t) { }
  virtual ~StringType() { }
  Kind kind() const { return String; }
  string name () const;
  string format(Value const & v) const;
  Value * parse(string const & v, string & remains) const;
  Value * create_default_value() const;
  double identify(string const & v) const;
  void autosetup(string_tools::strings const & learning);

  virtual DataType* clone() const { return new StringType(*this); }
private:
  // Here, we have information about quoting
  // strings, and how to escape strings
  void assign(DataType const & d) { }
};

#endif // _INC_DATATYPE_H

