// This is -*- C++ -*-
// $Id: RealSet.cpp,v 1.2 1998/12/30 02:21:43 trow Exp $

/* 
 * RealSet.cpp
 *
 * Copyright (C) 1998 EMC Capital Management, Inc.
 *
 * Developed by Jon Trowbridge <trow@emccta.com>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <config.h>
#include <string.h>
#include <stdlib.h>
#include <iostream>
#include <algorithm>
#include <functional>
#include "RealSet.h"

RealSet::RealSet() : epsilon_(1e-8), type_(0)
{
  clear();
}

RealSet::RealSet(const RealSet& ds)
{
  copy(ds);
}

RealSet&
RealSet::operator=(const RealSet& ds)
{
  copy(ds);
  return *this;
}

const double*
RealSet::sorted_data() const
{
  if (size() == 0) return 0; 
  if (sdata_.size()) return sdata_.data();

  RCArray<double> sd = data_;
  double* s = sd.data();

  if (sortedness() == 0) {
    std::sort(s, s+sd.size(), less<double>());
  } else if (sortedness() == -1) {
    for(size_t i=0; i<size()/2; ++i) {
      double t = s[i];
      s[i] = s[size()-1-i];
      s[size()-1-i] = t;
    }
  }
    
  // Sometimes you just have to fool the compiler...
  ((RealSet*)this)->sdata_ = sd;

  return (const double*)s;
}

double 
RealSet::min() const { 
  if (size() == 0)
    throw Exception("Dataset is empty");
  return min_; 
}
double 
RealSet::max() const { 
  if (size() == 0)
    throw Exception("Dataset is empty");
  return max_; 
}

double 
RealSet::range() const { return max() - min(); }

double
RealSet:: sum() const { return sum_; }

double
RealSet:: mean() const { 
  if (size() == 0)
    throw Exception("Dataset is empty");
  return sum_/size();
}

double
RealSet:: var() const {
  if (size() == 0)
    throw Exception("Dataset is empty");
  return sumsq_/size();
}

double
RealSet:: sdev() const {
  if (size() == 0)
    throw Exception("Dataset is empty");
  return sqrt(sumsq_/size());
}


double
RealSet:: sdevs() const { 
  if (size() < 2)
    throw Exception("Need at least two values");
  return sqrt(sumsq_/(size()-1));
}

double
RealSet::percentile(double p) const {
  const double* s = sorted_data();
  double t = p*(size()-1);
  int i = (int) t;
  if (s)
    return (i+1-t)*s[i] + (t-i)*s[i+1];
  else
    throw Exception("Dataset is empty");
}

double
RealSet:: med() const { return percentile(0.5); }
double
RealSet:: q1() const { return percentile(0.25); }
double
RealSet:: q3() const { return percentile(0.75); }
double
RealSet:: iqr() const { return q3() - q1(); } // interquartile range

double RealSet::decile(int i) const { 
    if (i<1||i>9) {
      throw Exception("Decile out of range");
    } else {
      return percentile(i/10.0);
    }
  }


double
RealSet::mean_trimmed(size_t l, size_t r) const
{
  if (l+r >= size())
    throw Exception("Not enough data");

  const double* s = sorted_data();
  double tot = sum();
  for(size_t i=0; i<l; ++i)
    tot -= s[i];
  for(size_t i=0; i<r; ++i)
    tot -= s[size()-1-i];
  return tot/(size()-l-r);
}

double
RealSet::mean_winsorized(size_t l, size_t r) const
{
  if (l+r >= size())
    throw Exception("Not enough data");

  if (size() == 0) {
    throw Exception("Dataset is empty");
  }

  const double* s = sorted_data();
  double tot = sum();
  for(size_t i=0; i<l; ++i)
    tot -= s[i];
  tot += l*s[l];
  for(size_t i=0; i<r; ++i)
    tot -= s[size()-1-i];
  tot += r*s[size()-1-r];
  return tot/size();
}

double
RealSet::moment(unsigned k, double x) const
{
  if (size() == 0) 
    throw Exception("Dataset is empty");

  const double* d = data();
  double sum=0;
  for(size_t i=0; i<size(); ++i)
    sum += pow(d[i]-x,k);
  return sum/size();
}

double
RealSet::gmean() const
{
  if (size() == 0)
    throw Exception("Dataset is empty");

  if (!is_positive(min()))
    throw Exception("Minimum is not positive");

  const double* d = data();
  double logsum=0;
  for(size_t i=0; i<size(); ++i)
    logsum += log(d[i]);
  return exp(logsum/size());
}

double
RealSet::hmean() const
{
  if (size() == 0) 
    throw Exception("Dataset is empty");

  double invsum=0;
  const double* d = data();
  for(size_t i=0; i<size(); ++i) {
    if (is_zero(d[i])) 
      throw Exception("Value is zero");
    invsum += 1/d[i];
  }
  if (is_zero(invsum))
    throw Exception("Inverse sum is zero");
  return size()/invsum;
}

double
RealSet::rms() const
{
  // "My work on free software is motivated by an idealistic goal: spreading
  // freedom and cooperation.  I want to encourage free software to spread,
  // replacing proprietary software which forbids cooperation, and thus
  // make our society better." -rms
  if (size() == 0) 
    throw Exception("Dataset is empty");
  const double* d = data();
  double sum=0;
  for(size_t i=0; i<size(); ++i)
    sum += d[i]*d[i];
  return sqrt(sum/size());
}

double
RealSet::meandev() const
{
  if (size() == 0) 
    throw Exception("Dataset is empty");
  const double* d = data();
  double sum=0, m=mean();
  for(size_t i=0; i<size(); ++i)
    sum += fabs(d[i]-m);
  return sum/size();
}

double
RealSet::meddev() const
{
  // This is expensive, and it duplicates code.  And it could be optimized.
  // Yipee!
  if (size() == 0) 
    throw Exception("Dataset is empty");
  double m=med();
  const double* dd = data();
  double* d = new double[size()];
  for(size_t i=0; i<size(); ++i)
    d[i] = fabs(dd[i]-m);
  std::sort(d,d+size(),less<double>());
  double md = size()%2 ? d[size()/2] : (d[size()/2-1]+d[size()/2])/2;
  delete [] d;
  return md;
}

double
RealSet::kurtosis() const
{
  if (size() == 0) 
    throw Exception("Dataset is empty");
  const double* d = data();
  double m=mean(), mom2=0, mom4=0;
  for(size_t i=0; i<size(); ++i) {
    double x = d[i]-m;
    x *= x;
    mom2 += x;
    mom4 += x*x;
  }
  return mom4/(mom2*mom2);
}

double
RealSet::skewness() const
{
  if (size() == 0) 
    throw Exception("Dataset is empty");
  const double* d = data();
  double m=mean(), mom2=0, mom3=0;
  for(size_t i=0; i<size(); ++i) {
    double x = d[i]-m;
    mom2 += x*x;
    mom3 += x*x*x;
  }
  return mom3/pow(mom2,1.5);
}

double 
RealSet::excess_kurtosis() const {
  return kurtosis()-3;
}

double 
RealSet::momental_skewness() const {
  return skewness()/2;
}

double
RealSet::durbin_watson() const
{
  size_t N = size();
  if (N < 2)
    throw Exception("Need at least two values");

  double run = 0;
  double run_sq = 0;
  const double* d = data();
  run_sq = d[0] * d[0];
  for(size_t i=1; i<N; ++i) {
    double x = d[i] - d[i-1];
    run += x*x;
    run_sq += d[i] * d[i];
  }

  return run / run_sq;
}

/*
  This is the test from pg. 93 of Madansky's
  "Prescriptions for Working Statisticians". 
  It checks for independence, with the alternative being
  AR(1) autoregression.  Does it have a name?
  It looks like a Durbin-Watson statistic with a slight
  change in the denominator.
*/
double
RealSet::AR1_independence_z() const
{
  size_t N = size();
  if (N < 2)
    throw Exception("Need at least two values");

  double run=0;
  const double* d = data();
  for(size_t i=1; i<N; ++i) {
    double x = d[i] - d[i-1];
    run += x*x;
  }

  double v = run / sumsq_;
  const double exp_mean = 2.0;
  const double exp_var = 4.0 * (N - 1)/(N*N - 1);

  return (v-exp_mean)/sqrt(exp_var);
}

double
RealSet::autocorr(size_t lag) const
{
  size_t N = size();
  if (lag >= N)
    throw Exception("Lag is too large");

  const double* d = data();
  const double m = mean();
  double run = 0;
  for(size_t i=lag; i<N; ++i)
    run += (d[i]-m)*(d[i-lag]-m);
  
  return (run/(N-lag)) / var();
}

double
RealSet::autocorr_z(size_t lag) const
{
  size_t N = size();
  const double ac = autocorr(lag);
  const double exp_mean = -1.0/(N-1);
  const double exp_var  = 1.0/N;
  return (ac - exp_mean) / sqrt(exp_var);
}

double
RealSet::covar(const RealSet& ds) const
{
  if (size() != ds.size()) 
    throw Exception("Datasets are not of the same size");
  if (size() == 0)
    throw Exception("Dataset is empty");

  const double* d = data();
  const double* dd = ds.data();
  const size_t N = size();
  double sumx=d[0], sumy=dd[0], Sxy=0;
  for(size_t i=1; i<N; ++i) {
    double x = d[i];
    double y = dd[i];
    sumx += x;
    Sxy += (x - sumx/(i+1))*(y - sumy/i);
    sumy += y;
    // Exercise for the reader: Why does this give us the right answer?
  }
  return Sxy/N;
}

double 
RealSet::corr(const RealSet& ds) const 
{ 
  double sdds = sdev() * ds.sdev();
  if (sdds == 0)
    throw Exception("Value not defined");
  return covar(ds)/sdds; 
}

double
RealSet::pooled_mean(const RealSet& ds) const
{
  if (ds.size() == 0 && size() == 0)
    throw Exception("Both datasets are empty");
  return (sum() + ds.sum()) / (size() + ds.size());
}

double
RealSet::pooled_var(const RealSet& ds) const
{
  if (ds.size() == 0 && size() == 0)
    throw Exception("Both datasets are empty");
  else if (ds.size() == 0)
    return var();
  else if (size() == 0)
    return ds.var();

  const RealSet* a = size() < ds.size() ? &ds : this;
  const RealSet* b = size() < ds.size() ? this : &ds;

  double sumsq = a->sumsq_;
  double sum = a->sum_;
  size_t count = a->size();

  const size_t N = b->size();
  for(size_t i=0; i<N; ++i) {
    double x = b->data(i);
    sum += x;
    sumsq += (x-sum/(count+1))*(x-(sum-x)/count);
    ++count;
  }

  return sumsq / count;
}

double
RealSet::weighted_mean(const RealSet& wt) const
{
  const size_t N = size();
  if (N == 0) 
    throw Exception("Dataset is empty");
  if (N != wt.size())
    throw Exception("Datasets are not of the same size");

  const double* d = data();
  const double* w = wt.data();
  double sum = 0.0;
  for (size_t i=0; i<N; ++i)
    sum += d[i] * w[i];

  return sum / wt.sum();
}

size_t
RealSet::greater_than(double x) const
{
  if (size() == 0 || max() <= x) return 0;
  if (min() > x) return size();

  const double* d = sorted_data();
  size_t i=0, j=size()-1;

  for(;;) {
    size_t mid=(i+j)/2;
    if (mid == size()-1) --mid;
    if (d[mid] <= x && x < d[mid+1])
      return size()-mid-1;
    else if (d[mid] > x) 
      j = mid;
    else
      i = mid;
  }
}

size_t
RealSet::less_than(double x) const
{
  if (size() == 0 || min() >= x) return 0;
  if (max() < x) return size();

  const double* d = data();
  size_t i=0, j=size()-1;

  for(;;) {
    size_t mid=(i+j)/2;
    if (mid == size()-1) --mid;
    if (d[mid] < x && x <= d[mid+1])
      return mid+1;
    else if (d[mid] >= x)
      j = mid;
    else
      i = mid;
  }
}

void
RealSet::reserve(size_t R)
{
  data_.reserve(R);
}

void
RealSet::add(double x)
{
  update_add(x);
  data_.add(x);
  if (sortedness_ > 0)
    sdata_ = data_;
  else
    sdata_.clear();
}

void
RealSet::add_at(size_t i, double x)
{
  if (i > size())
    throw Exception("Index out of range");

  update_add(x);
  data_.add_at(i,x);
  recalc_sortedness(); // sortedness will be wrong from update_add().
  if (sortedness_ > 0)
    sdata_ = data_;
  else
    sdata_.clear();
}

void
RealSet::remove_at(size_t i)
{
  if (i > size())
    throw Exception("Index out of range");

  update_remove(data(i));
  data_.remove_at(i);

  // Remember, removing an element preserves sortedness.  The only possible
  // change is to bring an unsorted list into a sorted state.
  if (sortedness_ == 0)
    recalc_sortedness();

  if (sortedness_ > 0)
    sdata_ = data_;
  else
    sdata_.clear();
}

void
RealSet::add(double* xs, size_t count)
{
  data_.add(xs,count);
  recalc();
}
void
RealSet::add_at(size_t i, double* xs, size_t count)
{
  if (i > size())
    throw Exception("Index out of range");
  else if (i == size())
    add(xs, count);
  else {
    data_.add_at(i,xs,count);
    recalc();
  }
}

void
RealSet::remove_at(size_t i, size_t j)
{
  if (i > j) {
    throw Exception("Illegal range");
  }
  if (j >= size()) {
    throw Exception("Index out of range");
  }
  data_.remove_at(i,j);
  recalc();
}

void
RealSet::set(size_t i, double x)
{
  if (i >= size()) {
    throw Exception("Index out of range");
  }
  data_.data()[i] = x;
  recalc();
}

void
RealSet::set(size_t i, const double* xs, size_t count)
{
  if (i + count >= size()) {
    throw Exception("Index out of range");
  }
  double* array = data_.data();
  size_t j = 0;
  while (j < count) {
    array[i+j] = xs[j];
    ++j;
  }
  recalc();
}

void
RealSet::clear()
{
  data_.clear();
  sdata_.clear();
  sum_ = sumsq_ = 0;
  sortedness_ = 2; // constant
}

void
RealSet::append(const RealSet& ds)
{
  // clearly, this could be optimized...
  for(size_t i=0; i<ds.size(); ++i)
    add(ds.data(i));
}

/*
void
RealSet::apply(double (*fn)(double))
{
  double* d = data_.data();
  for(size_t i=0; i<size(); ++i)
    d[i] = fn(d[i]);
  recalc();
}
*/

void
RealSet::linear_transform(double a, double b)
{
  double* d = data_.data();

  for(size_t i=0; i<size(); ++i) 
    d[i] = a*d[i]+b;

  
  double m1 = a*min_+b, m2 = a*max_+b;
  min_ = m1<m2 ? m1 : m2;
  max_ = m1<m2 ? m2 : m1;

  sum_ = a*sum_+b*size();
  sumsq_ *= a*a;
  if (a<0 && sortedness_ != 2)
    sortedness_ = -sortedness_;

  if (sdata_.size()) {
    double* s = sdata_.data();
    // linearly transform our sdata buffer
    for(size_t i=0; i<size(); ++i)
      s[i] = a*s[i]+b;

    // if a<0, reverse the sdata to preserve sortedness
    if (a<0) {
      for(size_t i=0; i<size()/2; ++i) {
	double t=s[i];
	s[i]=s[size()-1-i];
	s[size()-1-i]=t;
	}
    }
  }
}

void
RealSet::log_transform()
{
  if (min() <= 0) {
    clear();
    return;
  }
  double* d = data_.data();
  for (size_t i=0; i<size(); ++i)
    d[i] = log(d[i]);


  if (sdata_.size()) {
    double* s = sdata_.data();
    for(size_t i=0; i<size(); ++i)
      s[i] = log(s[i]);
  }

  recalc_saving_sorted();
}

void
RealSet::exp_transform()
{
  double* d = data_.data();
  for (size_t i=0; i<size(); ++i)
    d[i] = exp(d[i]);


  if (sdata_.size()) {
    double* s = sdata_.data();
    for(size_t i=0; i<size(); ++i)
      s[i] = exp(s[i]);
  }

  recalc_saving_sorted();
}

void
RealSet::logit_transform()
{
  if (min() <= 0 || max() >= 1) {
    clear();
    return;
  }
  double* d = data_.data();
  for(size_t i=0; i<size(); ++i)
    d[i] = log(d[i]/(1-d[i]));

  // the logit transform is order-preserving, so we can zap our
  // sorted data as well
  if (sdata_.size()) {
    double* s = sdata_.data();
    for(size_t i=0; i<size(); ++i)
      s[i] = log(s[i]/(1-s[i]));
  }

  recalc_saving_sorted();
}

void
RealSet::reverse()
{
  if (sortedness_ == 2) return;

  double* d = data_.data();

  for(size_t i=0; i<size()/2; ++i) {
    double t = d[i];
    d[i] = d[size()-i-1];
    d[size()-1-i] = t;
  }
  sortedness_ = -sortedness_;
}

void
RealSet::sort(int dir)
{
  // If we are already sorted properly, ignore
  if (sortedness_ == 2 || 
      (sortedness_ == 1 && dir >= 0) ||
      (sortedness_ == -1 && dir < 0))
    return;

  // If we have sorted data already, use it!
  if (dir >= 0 && sdata_.size()) {
    data_ = sdata_;
    sortedness_ = 1;
    return;
  }

  // If we are being asked to "opposite-sort" already-sorted data, do so
  // via reverse()
  if ((dir >= 0 && sortedness_ == -1) || (dir < 0 && sortedness_ == 1)) {
    reverse();
  } else if (dir >= 0) {
    // Otherwise there is no avoiding it, so just sort the data...
    double* d = data_.data();
    std::sort(d, d+size(), less<double>());
    sortedness_ = 1;
  } else {
    double* d = data_.data();
    std::sort(d, d+size(), greater<double>());
    sortedness_ = -1;
  }

  if (sortedness_ > 0)
    sdata_ = data_;
}

void
RealSet::scramble()
{
  if (size() == 0 || sortedness_ == 2) return;

  // If we were sorted before, copy the sorted data so we don't lose it
  double* d = data_.data();

  for(size_t i=0; i<size()-1; ++i) {
    size_t j = random().random_integer(i,size()-1);
    if (i != j) {
      double t = d[i];
      d[i] = d[j];
      d[j] = t;
    }
  }

  recalc_sortedness();
}

void
RealSet::permute(const Permutation& p)
{
  if (size() != p.size())
    throw Exception("Permutation and data are not of the same size");
  
  double* d = data_.data();
  p.permute(d);

  recalc_sortedness();
}

void
RealSet::rank()
{
  pair<double,int>* rankdata = new pair<double,int>[size()];
  double* d = data_.data();

  for(size_t i1=0; i1<size(); ++i1) {
    rankdata[i1].first = d[i1];
    rankdata[i1].second = i1;
  }

  std::sort(rankdata, rankdata+size(), less< pair<double,int> >());

  // set ranks, handling ties by the usual convention
  unsigned count=1;
  for(size_t i=0; i<size(); /* increment down below */ ) {
    size_t j=i+1;
    unsigned count2=0;
    while (j < size() && are_equal(rankdata[i].first, rankdata[j].first)) {
      ++j;
      ++count2;
    }
    while (i != j)
      rankdata[i++].first = (2*count + count2)/2.0; // we increment i here
    count += 1+count2;
  }

  for(size_t i=0; i<size(); ++i)
    d[rankdata[i].second] = rankdata[i].first;

  recalc(); // This could be optimized...

  delete [] rankdata;
}

Permutation
RealSet::sorting_permutation(int dir) const
{
  Permutation p;

  if (sortedness_ == 2 ||
      (dir >= 0 && sortedness_ == 1) ||
      (dir < 0 && sortedness_ == -1)) {

    p.set_identity(size());

  } else if ((dir > 0 && sortedness_ == -1) ||
	     (dir < 0 && sortedness_ == 1)) {

    p.set_reverse(size());

  } else {

    pair<double,size_t>* tmp = new pair<double,size_t>[size()];
    size_t* tmp2 = new size_t[size()];
    const double* d = data_.data();
    for (size_t i=0; i<size(); ++i) {
      tmp[i].first = d[i];
      tmp[i].second = i;
    }

    if (dir >= 0) 
      std::sort(tmp, tmp+size(), less< pair<double,size_t> >());
    else
      std::sort(tmp, tmp+size(), greater< pair<double,size_t> >());

    for(size_t i=0; i<size(); ++i)
      tmp2[tmp[i].second] = i;

    p.initialize(size(), tmp2);
    delete [] tmp;
    delete [] tmp2;
  }

  return p;
}

void
RealSet::update_add(double x)
{
  // update statistics /////

  size_t N = size();

  if (N == 0 || x < min_)
    min_ = x;
  if (N == 0 || x > max_)
    max_ = x;

  sum_ += x;
  if (N > 0) {
    sumsq_ += (x-sum_/(N+1))*(x-(sum_-x)/N); // cool, huh?

    // Update sortedness flag
    if (sortedness_) {
      const double* d = data_.data_const();
      double prev = d[N-1];
      if (sortedness_ == 2) { // constant
	if (prev < x)
	  sortedness_ = 1;
	else if (prev > x)
	  sortedness_ = -1;
      } else if (sortedness_ == 1) { // ascending
	if (prev > x)
	  sortedness_ = 0;
      } else if (sortedness_ == -1) { // descending
	if (prev < x)
	  sortedness_ = 0;
      }
    }
  }
}

void
RealSet::update_remove(double x)
{
  size_t N = size()-1;
  if (N > 0)
    sumsq_ -= (x-sum_/(N+1))*(x-(sum_-x)/N);
  sum_ -= x;

  if (N != 0) {
    const double* d = data_.data_const();
    size_t NN = size();
    if (min_ == x) {
      bool skip_x = true;
      min_ = d[0];
      if (x == d[0]) {
	min_ = d[1];
	skip_x = false;
      }
      for(size_t i=1; i<NN; ++i) {
	if (d[i] == x && skip_x) {
	  skip_x = false;
	  continue;
	}
	if (min_ > d[i])
	  min_ = d[i];
      }
    }
    if (max_ == x) {
      bool skip_x = true;
      max_ = d[0];
      if (x == d[0]) {
	max_ = d[1];
	skip_x = false;
      }
      for(size_t i=1; i<NN; ++i) {
	if (d[i] == x && skip_x) {
	  skip_x = false;
	  continue;
	}
	if (max_ < d[i])
	  max_ = d[i];
      }
    }
  }
  // Leaves sortedness in fubared state.
}

// There is some code duplication between these and add(), but calling a
// recalc() should be a lot more efficient than just repeatedly add()-ing
// the numbers if you've been fiddling directly with data_.

void
RealSet::recalc_sortedness()
{
  sortedness_ = 2;
  const double* d = data_.data_const();
  size_t N = data_.size();

  for(size_t i=1; i<N && sortedness_; ++i) {
    double x = d[i];
    double prev = d[i-1];
    if (sortedness_ == 2) {
      if (prev < x)
	sortedness_ = 1;
      else if (prev > x)
	sortedness_ = -1;
    } else if (sortedness_ == 1) {
      if (prev > x)
	sortedness_ = 0;
    } else if (sortedness_ == -1) {
      if (prev < x)
	sortedness_ = 0;
    }
  }
}

void
RealSet::recalc()
{
  recalc_saving_sorted();
  sdata_.clear();
}

void
RealSet::recalc_saving_sorted()
{
  sum_ = sumsq_ = 0;
  sortedness_ = 2;
  if (size() == 0) return;

  const double* d = data_.data_const();
  size_t N = data_.size();

  min_ = max_ = d[0];
  for (size_t i=0; i<N; ++i) {
    double x = d[i];
    if (min_ > x) min_ = x;
    if (max_ < x) max_ = x;
    sum_ += x;
    if (i>0) {
      sumsq_ += (x-sum_/(N+1))*(x-(sum()-x)/N);
      if (sortedness_) {
	double prev = d[i-1];
	if (sortedness_ == 2) {
	  if (prev < x)
	    sortedness_ = 1;
	  else if (prev > x)
	    sortedness_ = -1;
	} else if (sortedness_ == 1) {
	  if (prev > x)
	    sortedness_ = 0;
	} else if (sortedness_ == -1) {
	  if (prev < x)
	    sortedness_ = 0;
	}
      }
    }
  }

  sdata_.clear();
}

void
RealSet::copy(const RealSet& ds)
{
  if (this == &ds) return; // can't copy ourselves.

  clear();
  label_ = ds.label_;
  data_ = ds.data_;
  sdata_ = ds.sdata_;
  min_ = ds.min_;
  max_ = ds.max_;
  sum_ = ds.sum_;
  sumsq_ = ds.sumsq_;
  sortedness_ = ds.sortedness_;
  epsilon_ = ds.epsilon_;
}



// $Id: RealSet.cpp,v 1.2 1998/12/30 02:21:43 trow Exp $

