// This is -*- C++ -*-
// $Id: RealSet.h,v 1.2 1998/12/30 02:21:44 trow Exp $

/* 
 * RealSet.h
 *
 * Copyright (C) 1998 EMC Capital Management, Inc.
 *
 * Developed by Jon Trowbridge <trow@emccta.com>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#ifndef _INC_DATASET_H
#define _INC_DATASET_H

#include <iostream>
#include <math.h>
#include <string>
#include "RCArray.h"
#include "DataType.h"
#include "Permutation.h"
#include "Exception.h"
#include "DataSet.h"
#include "Access.h"
#include "Value.h"

using namespace std;

class RealSet : public DataSet {
public:
  RealSet();
  RealSet(const RealSet&);
  ~RealSet() { }

  RealSet& operator=(const RealSet&);

  size_t size() const { return data_.size(); }

  Value const * operator[](size_t n) const {
    return ValueFactory::factory()->create(data(n));
  }

  // make sure we have allocated room for at least R entries
  void reserve(size_t R); 

  void push_back(Value const * v) {
    add(Access::get_real(*v));
  }

  void insert(size_t pos, Value const * v) {
    add_at(pos, Access::get_real(*v));
  }

  void pop_back() {
    if (!empty()) {
      remove_at(size() - 1);
    }
  }

  void erase(size_t pos) {
    remove_at(pos);
  }

  void erase(size_t first, size_t last) {
    remove_at(first, last);
  }

  void clear();

  // Return a permutation that will, when applied, bring the RealSet into
  // sorted order.  (Does not sort the RealSet, just produces instructions
  // for how to do so!)
  Permutation sorting_permutation(int dir = 1) const;

  void permute(const Permutation&);

  // useful for debugging, too gross for much else
  void spew(ostream& out, char sep = ' ') const {
    const double* d = data();
    for(size_t i=0; i<size(); ++i) out << d[i] << sep;
    if (sep != '\n')
      out << endl;
  }


  // RealSet specific methods

  const double* data() const { return data_.data(); }

  // note: no bounds checking
  double data(size_t i) const { return (data_.data())[i]; } 

  const double* sorted_data() const;
  double sorted_data(size_t i) const { return sorted_data()[i]; }

  double min() const;
  double max() const;
  double range() const;

  double sum() const;
  double mean() const;
  double var() const;
  double sdev() const;
  double sdevs() const;

  double percentile(double p) const;
  double med() const;
  double q1() const;
  double q3() const;
  double iqr() const; // interquartile range
  double decile(int i) const;

  // There are some computed (non-cached) descriptive statistics.  They are
  // a bit more obscure, so I don't bother keeping a running tally of them
  // via add().  (Or, in some cases, I can't...)
  double mean_trimmed(size_t left, size_t right) const;
  double mean_trimmed(size_t t) const { return mean_trimmed(t,t); }
  double mean_winsorized(size_t left, size_t right) const;
  double mean_winsorized(size_t t) const { return mean_winsorized(t,t); }
  double moment(unsigned k, double x) const; // k-th moment about x
  double moment(unsigned k) const { return moment(k,mean()); }
  double gmean() const;    // geometric mean
  double hmean() const;    // harmonic mean
  double rms() const;      // root mean square
  double meandev() const;  // mean deviation
  double meddev() const;   // median deviation
  double kurtosis() const; // kurtosis
  double skewness() const; // skewness

  double excess_kurtosis() const;
  double momental_skewness() const;

  double durbin_watson() const;
  double AR1_independence_z() const;
  double autocorr(size_t lag) const;
  double autocorr_z(size_t lag) const;

  // Here are some functions of two RealSets
  double covar(const RealSet& ds) const;
  double corr(const RealSet& ds) const;
  double pooled_mean(const RealSet& ds) const;
  double pooled_var(const RealSet& ds) const;
  double weighted_mean(const RealSet& wt) const;

  int sortedness() const { return sortedness_; }
  bool increasing() const { return sortedness_ == 1; }
  bool decreasing() const { return sortedness_ == -1; }
  bool unordered() const { return sortedness_ == 0; }
  bool constant() const { return sortedness_ == 2; }

  size_t greater_than(double x) const;
  size_t less_than(double x) const ;
  size_t between(double a, double b) const {
    return size()-less_than(a)-greater_than(b);
  }
  size_t equal_to(double x) const {
    return size()-less_than(x)-greater_than(x);
  }

  void add(double);
  void add_at(size_t i, double);
  void remove_at(size_t i);

  void add(double*, size_t count);
  void add_at(size_t i, double*, size_t count);
  void remove_at(size_t i, size_t j);

  void set(size_t i, double);
  void set(size_t i, const double*, size_t count);

  void append(const RealSet&);

//  This case is covered by the template, and thus ambigious... (Asger)
//  void apply(double (*)(double));

  template<class FuncObj>
  void apply(FuncObj f) {
    for(size_t i=0; i<size(); ++i)
      (data_.data())[i] = f((data_.data())[i]);
    recalc();
  }

  // transform by x -> ax+b
  void linear_transform(double a, double b);
  // transform by x -> log(x) (natural log, of course...)
  void log_transform();
  void exp_transform();
  void logit_transform();

  void reverse();
  void sort(int dir = 1);
  void scramble();
  void rank();

  size_t random_index() const { 
    return (size_t)random().random_integer(0,size()-1);
  }
  double random_element() const { return data(random_index()); }

  double epsilon() const { return epsilon_; }
  void set_epsilon(double e) { epsilon_ = e; }

private:
  void update_add(double);
  void update_remove(double);
  void recalc_sortedness();
  void recalc();
  void recalc_saving_sorted();
  void copy(const RealSet&);

  // convenience routines...
  bool is_zero(double x) const { return fabs(x) < epsilon_; }
  bool are_equal(double x, double y) const { return is_zero(x-y); }
  bool is_positive(double x) const { return x > epsilon_; }

  string label_;

  RCArray<double> data_;
  RCArray<double> sdata_;

  double min_, max_, sum_;
  // If we call our data values x_1,...,x_N, then sumsq_ contains
  // \sum_{i=0}^N(x_i-\mu)^2, where \mu is the mean of the x_i.
  double sumsq_;

  // sortedness == +1 if data is increasing
  //            == -1 if data is decreasing
  //            ==  0 if data is unordered
  //            == +2 if the data is constant (a bad hack)
  int sortedness_;

  double epsilon_;

  const DataType* type_;
};

#endif // _INC_DATASET_H

// $Id: RealSet.h,v 1.2 1998/12/30 02:21:44 trow Exp $
