// This is -*- C++ -*-
// $Id: Value.h,v 1.4 1998/12/30 23:14:29 hp Exp $

/* 
 * Value.h
 *
 * Copyright (C) 1998 Asger Alstrup Nielsen
 *
 * Developed by Asger Alstrup Nielsen <alstrup@diku.dk>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#ifndef _INC_VALUE_H
#define _INC_VALUE_H

#include <string>
#include "DataType.h"

using namespace std;

// Base class for values
class Value {
public:
  Value(Value const & v) { assign(v); }
  virtual ~Value() { if (type_ != 0) delete type_; }
  virtual Value & operator=(Value const & v) { assign(v); return *this; }
  virtual void set_value(string const & s);
  virtual string format() const;
  DataType const * type() const { return type_; }
  void set_type(DataType const & t);
  virtual Value* clone() const { return new Value(*this); }
protected:
  Value() : type_(0) { }
  Value(DataType const & t) : type_(t.clone()) { }
  void assign(Value const & v);
private:
  DataType * type_;
};

// A floating point value (represented as a double)
class Real : public Value {
public:
  Real();
  Real(Value const & v);
  Real(double d);
  void set_value(string const & s);
  string format() const;

  void set_real(double d);
  double get_real() const;

  virtual Value* clone() const { return new Real(*this); }
private:
  void assign(Value const & v);
  double value_;
};

// A string value
class String : public Value {
public:
  String() { };
  String(Value const & v) { assign(v); }
  // This one does not parse the string
  String(string const & v) : Value(StringType()), value_(v) { }

  // This one parses the string
  void set_value(string const & s) { value_ = s; }
  // This one formats the string
  string format() const { return value_; }

  // This one does not parse the string
  void set_string(string & s) { s = value_; }
  string get_string() const { return value_; }

  virtual Value* clone() const { return new String(*this); }
private:
  void assign(Value const & v);
  string value_;
};


// A singleton factory class for creating values.
// Usage:
//   Value * v = ValueFactory::factory()->create(something);
// where something can be a double, string or any other
// supported representation type.
// The receiver has the responsibility to delete the returned
// object.
class ValueFactory {
public:
  static ValueFactory * factory() {
    if (factory_ == 0) {
      factory_ = new ValueFactory();
    }
    return factory_;
  }
  
  Real * create(double d) {
    Real * r = new Real();
    r->set_real(d);
    return r;
  }
  
  String * create(string const & s) {
    return new String(s);
  }

protected:
  ValueFactory() { }

private:
  static ValueFactory * factory_;
};

#endif
