// This is -*- C++ -*-
// $Id: bootstrap.cpp,v 1.15 1999/01/21 03:52:06 trow Exp $

/* bootstrap.cpp
 *
 * Copyright (C) 1998 EMC Capital Management, Inc.
 *
 * Developed by Jon Trowbridge <trow@emccta.com>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <config.h>
#include "bootstrap.h"
#ifdef HAVE_LIBPTHREAD
#include <pthread.h>
#endif

struct bootargs {
  size_t N;
  const RealSet* source;
  RealSet* target;
};

RealSet
call_inner_bootstrap(void* (fn)(void*), size_t N, const RealSet& ds,
		     int threads)
{
#ifdef HAVE_LIBPTHREAD
  if (threads > 1) {
    RealSet* targets = new RealSet[threads];
    bootargs* args = new bootargs[threads];
    pthread_t* th_ids = new pthread_t[threads];

    // Spawn the appropriate number of threads, assigning them each an
    // appropriate chunk of the job.
    size_t remaining = N;
    size_t chunk = N/threads;
    if (chunk == 0) chunk = 1;
    for(int i=0; i<threads; ++i) {

      if (chunk > remaining || i == threads-1)
	chunk = remaining;

      args[i].N = chunk;
      args[i].target = &(targets[i]);
      args[i].source = &ds;

      remaining -= chunk;

      pthread_create(&(th_ids[i]), NULL, fn,
		     (void*)&(args[i]));
    }

    // As each thread finishes, stuff its results into boot.
    RealSet boot;
    for(int i=0; i<threads; ++i) {
      pthread_join(th_ids[i], NULL);
      boot.append(targets[i]);
    }

    delete [] targets;
    delete [] args;
    delete [] th_ids;

    return boot;
  }
#endif
  RealSet boot;
  bootargs args = { N, &ds, &boot };
  fn(&args);
  return boot;
}

/// /// /// /// /// /// /// /// /// /// /// /// /// /// /// /// /// /// ///

void*
inner_bootstrap_mean(void* varg)
{
  bootargs* b = (bootargs*)varg;

  size_t N = b->N;
  if (N == 0) return 0;

  const RealSet* src = b->source;
  RealSet* tgt = b->target;

  size_t dsN = src->size();
  const double* data = src->data();

  tgt->reserve(N);

  Random rnd;
  int* ind = new int[dsN];

  for(size_t i=0; i<N; ++i) {
    double sum=0;
    rnd.random_integer(0, dsN-1, ind, dsN);
    for(size_t j=0; j<dsN; ++j)
      sum += data[ind[j]];
    tgt->add(sum / dsN);
  }

  delete [] ind;

  return 0;
}

RealSet
bootstrap_mean(size_t N, const RealSet& ds, int threads)
{
  return call_inner_bootstrap(inner_bootstrap_mean, N, ds, threads);
}

/// /// /// /// /// /// /// /// /// /// /// /// /// /// /// /// /// /// ///

void*
inner_bootstrap_median(void* varg)
{
  bootargs* b = (bootargs*)varg;

  size_t N = b->N;
  if (N == 0) return 0;

  const RealSet* src = b->source;
  RealSet* tgt = b->target;

  size_t dsN = src->size();
  const double* sdata = src->sorted_data();

  tgt->reserve(N);

  Random rnd;
  int* ind = new int[dsN];
  int* count = new int[dsN];

  for(size_t i=0; i<N; ++i) {
    rnd.random_integer(0, dsN-1, ind, dsN); // pre-generate indices
    memset(count, 0, dsN*sizeof(int));      // clear the count buffer
    
    for(size_t j=0; j<dsN; ++j)
      ++count[ind[j]];

    int run=0, target=(dsN-1)/2;
    size_t j=0;
    for(;j<dsN && run <= target; ++j)
      run += count[j];

    double med;
    if (dsN % 2 == 0 /* even */) {
      size_t jj=j-1;
      if(run - target == 1)
	while (count[++jj] == 0);
      med = (sdata[j-1] + sdata[jj])/2;
    } else {
      med = sdata[j-1];
    }
    tgt->add(med);
  }

  delete [] ind;
  delete [] count;

  return 0;
}

RealSet
bootstrap_median(size_t N, const RealSet& ds, int threads)
{
  // Make sure the sorted data exists, since RealSet isn't thread safe (yet?)
  // We do the copy and this non-optimal sort (we lose the sort if ds wasn't
  // already sorted, which kinda sucks) just to make sure that the 
  // "transparent" sort doesn't get optimized away by an overzealous compiler
  // that believes our web of lies involving const.
  RealSet our_ds = ds;
  our_ds.sort();
  return call_inner_bootstrap(inner_bootstrap_median, N, our_ds, threads);
}

/// /// /// /// /// /// /// /// /// /// /// /// /// /// /// /// /// /// ///

void*
inner_bootstrap_sdev(void* varg)
{
  bootargs* b = (bootargs*)varg;

  size_t N = b->N;
  if (N == 0) return 0;

  const RealSet* src = b->source;
  RealSet* tgt = b->target;

  size_t dsN = src->size();
  const double* data = src->data();

  tgt->reserve(N);

  Random rnd;
  int* ind = new int[dsN];

  for(size_t i=0; i<N; ++i) {
    rnd.random_integer(0, dsN-1, ind, dsN);
    double sum=data[ind[0]], sumsq=0;
    for(size_t j=0; j<dsN; ++j) {
      double x = data[ind[j]];
      sum += x;
      sumsq += (x-sum/(j+1))*(x-(sum-x)/j);
    }
    tgt->add(sqrt(sumsq/dsN));
  }

  delete [] ind;

  return 0;
}

RealSet
bootstrap_sdev(size_t N, const RealSet& ds, int threads)
{
  return call_inner_bootstrap(inner_bootstrap_sdev, N, ds, threads);
}

/// /// /// /// /// /// /// /// /// /// /// /// /// /// /// /// /// /// ///

void
bootstrap_linreg_pairs(size_t N, const RealSet& x, const RealSet& y,
		       RealSet& slope, RealSet& inter, RealSet& r)
{
  if (x.size() != y.size()) {
    throw Exception("Datasets are not of the same size");
  }
  if (x.size() == 0) {
    throw Exception("Dataset is empty");
  }

  slope.clear();
  inter.clear();
  r.clear();

  slope.reserve(N);
  inter.reserve(N);
  r.reserve(N);

  const size_t dsN = x.size();
  const double* dx = x.data();
  const double* dy = y.data();

  int* ind = new int[dsN];

  Random rnd;
  for(size_t i=0; i<N; ++i) {
    rnd.random_integer(0, dsN-1, 0, dsN);
    double sumx=dx[0], sumy=dy[0], Sx=0, Sy=0, Sxy=0;
    double x,y,x0,y0;
    for(size_t j=1; j<dsN; ++j) {
      x = dx[j];
      y = dy[j];
      sumx += x;
      x0 = x - sumx/(j+1);
      y0 = y - sumy/j;
      sumy += y;
      Sx += x0 * (x - (sumx-x)/j);
      Sxy += x0 * y0;
      Sy += (y - sumy/(j+1)) * y0;
    }
    double m = Sxy/Sx;
    double b = sumy/dsN - m*sumx/dsN;
    slope.add(m);
    inter.add(b);
    r.add(b*sqrt(Sx/Sy));
  }

  delete [] ind;

}


// $Id: bootstrap.cpp,v 1.15 1999/01/21 03:52:06 trow Exp $
