// This is -*- C++ -*-
// $Id: stringtools.h,v 1.4 1998/12/14 18:48:29 alstrup Exp $

/* 
 * stringtools.h
 *
 * Copyright (C) 1998 Asger Alstrup Nielsen
 *
 * Developed by Asger Alstrup Nielsen <alstrup@diku.dk>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#ifndef _INC_STRINGTOOLS_H
#define _INC_STRINGTOOLS_H

#include <string>
#include <vector>
#include <stdio.h>

using namespace std;

/*
 * This is a struct that keeps a bunch of useful string manipulation routines.
 * Create a dummy instance of the struct to use it.
 */
template<class str>
struct basic_string_tools {
  // A vector of strings
  typedef vector<str> strings;

  /*
   * Trims characters off at both ends of the string.
   *
   * trim("  abc  ") == "abc" == trim("..abc..", ".");
   */
  str trim(str const & s, str const & trim_chars = " \t") {
    return cut(rcut(s, trim_chars), trim_chars);
  }

  /*
   * Cuts characters off at the front of the string.
   *
   * cut("  abc  ") == "abc  " == cut("..abc..", "abc..");
   */
  str cut(str const & s, str const & trim_chars = " \t") {
    if (trim_chars.empty() || s.empty() || s.find_first_of(trim_chars) == s.npos) return s;
    str work = s;
    size_t pos = work.find_first_not_of(trim_chars);
    if (pos == work.npos) {
      work.erase(work.begin(), work.end());
    } else if (pos != 0) {
      work.erase(0, pos);
    }
#if 0
    if (s != work) {
      cout << "cut '" << s.c_str() << "' with '" << trim_chars.c_str() << "'\n";
      cout << " -> '" << work.c_str() << "'\n";
    }
#endif
    return work;
  }

  /*
   * Cuts characters off at the end of the string.
   *
   * rcut("  abc  ") == "  abc" == rcut("..abc..", "..abc");
   */
  str rcut(str const & s, str const & trim_chars = " \t") {
    if (trim_chars.empty() || s.empty() || s.find_first_of(trim_chars) == s.npos) return s;
    str work = s;
    size_t pos = work.find_last_not_of(trim_chars);
    if (pos != work.npos) {
      work = substring(work, 0, pos);
    } else {
      // It seems that my STL library is buggy, so we need this:
      work = substring(work, 0, work.find_first_of(trim_chars) - 1);
      // rather than work.clear();
    }
#if 0
    if (s != work) {
      cout << "rcut '" << s.c_str() << "' with '" << trim_chars.c_str() << "'\n";
      cout << "  -> '" << work.c_str() << "'\n";
    }
#endif
    return work;
  }

  /*
   * Adds two strings such that it looks nice.
   *
   * concatenate_words("This", "is a test") == "This is a test";
   * concatenate_words("Empty", "") == "Empty";
   * concatenate_words("", "Empty") == "Empty";
   * concatenate_words("", "") == "";
   * concatenate_words(" This ", " is a test ") == "This is a test";
   */
  str concatenate_words(str const & s1, str const & s2) {
    str work = trim(s1);
    if (!work.empty()) {
      work += ' ';
    }
    work += cut(s2);
    return rcut(work);
  }

  /*
   * Convert an integer to a string
   *
   * int2str(10) = "10";
   */
  str int2str(int v) {
    char buf[10];
    sprintf(buf, "%d", v);
    return str(buf);
  }

  str int2str(unsigned int v) {
    char buf[10];
    sprintf(buf, "%u", v);
    return str(buf);
  }

  /*
   * Extract a piece of a string.
   * Contrary to the STL substr, this method takes two
   * indexes, rather than an index and a length.
   *
   * Also, indexes beyond the end of the string are handled
   * in a tolerant manner.
   */
  str substring(str const & s, size_t i1, size_t i2) {
    if (i1 >= s.length() || i1 > i2) {
      return str();
    }
    if (i2 >= s.length()) {
      i2 = s.length() - 1;
    }
    str result = s.substr(i1, (i2-i1) + 1);
#if 0
    if (result != s) {
      cout << "substring '" << s.c_str() << "' " << i1 << "-" << i2 << ": '" << result.c_str() << "'\n";
    }
#endif 
    return result;

  }

  /*
   * Split a string at a given delimiter.
   *
   * str head, tail;
   * split("This is a -!- test -!-", "-!-", head, tail);
   * head == "This is a " ; tail == " test -!-";
   * split("This is a -!- test -!-", "-!-", head, tail, true);
   * head == "This is a " ; tail == "-!- test -!-";
   * split("This is a test", "-!-", head, tail);
   * head == "This is a test" ; tail == "";
   */
  void split(str const & s, str const & delimiter, str & head, str & tail,
	     bool keep_delimiter_in_tail = false) {
    size_t pos = s.find(delimiter);
    if (pos == s.npos) {
      head = s;
      tail = str();
    } else {
      if (pos == 0) {
	head = string();
      } else {
	head = substring(s, 0,pos-1);
      }
      if (!keep_delimiter_in_tail) {
	pos += delimiter.length();
      }
      tail = substring(s, pos, s.length());
    }
  }


  /*
   * Does prefix match this string?
   */
  bool prefix(str const & s, str const & prefix) {
    return s.find(prefix) == 0;
  }
  
  // Expand tabs
  str expand_tabs(str const & line, size_t tab_size = 8) {
    str l = line;
    while(true) {
      size_t pos = l.find("\t");
      if (pos == l.npos)
	break;
      int amount = tab_size - (pos % tab_size);
      l.erase(pos, 1);
      l.insert(pos, amount, ' ');
    }
    return l;
  }

  // Is this for certain a character that's part of a number?
  bool is_number_char(str c) {
    return isdigit(c[0]);
  }

  // Could this character be part of a line of numbers,
  // but also is it also likely to be part of a line of headings?
  bool ambigious_number_char(str c) {
    return c[0] == '.' || c[0] == ',' 
           || c[0] == 'e' || c[0] == 'E' 
           || c[0] == '+' || c[0] == '-'
           || c[0] == ' ' || c[0] == '\t';
  }
};

typedef basic_string_tools<string> string_tools;
extern string_tools string_tool;

// typedef basic_string_tools<wstring> wstring_tools;
// extern wstring_tools wstring_tool;
#endif
